class QQError(RuntimeError):
    pass


class ApplicationNotStartedError(QQError):
    pass


class AlreadyStartedError(QQError):
    pass


class PluginLacksOfKeyError(QQError):
    pass


class WrongKeyForPluginError(QQError):
    pass


class PluginKeyAlreadyDefinedError(QQError):
    pass


class InjectorNotInicialized(QQError):
    """
    This error means the injector was not properly created. Probaly, used:

    def fun(param = Injector):
        pass

    Instead of:

    def fun(param = Injector()):
        pass
    """

    def __init__(self):
        super().__init__(self.__doc__)



from collections import OrderedDict

from qq.errors import PluginKeyAlreadyDefinedError
from qq.errors import PluginLacksOfKeyError
from qq.errors import WrongKeyForPluginError
from qq.types import Plugin
from qq.types import PluginKey


class PluginContainer(OrderedDict):
    def __call__(self, plugin: Plugin):
        if not plugin.key:
            raise PluginLacksOfKeyError()
        self[plugin.key] = plugin

    def __setitem__(self, key: PluginKey, plugin: Plugin):
        if plugin.key and plugin.key != key:
            raise WrongKeyForPluginError()
        if key in self:
            raise PluginKeyAlreadyDefinedError()
        super().__setitem__(key, plugin)



from typing import List

from celery import Celery

from qq.finder import ObjectFinder


class TaskFinder(ObjectFinder):
    def __init__(
        self,
        parent: str,
        ignore_list: List[str] = None,
        cache_key: str = None,
        celery_app: Celery = None,
    ):
        super().__init__(parent, ignore_list, cache_key)
        self.celery_app = celery_app

    def is_collectable(self, element: object):
        return getattr(element, "_app", None) == self.celery_app



from tornado.httpserver import HTTPServer
from tornado.log import enable_pretty_logging
from tornado.web import Application as Tornado

from qq.application import Application
from qq.plugins.settings import SettingsBasedPlugin

DEBUG_KEY = "debug"
SERVE_TRACEBACK_KEY = "serve_traceback"


class TornadoPlugin(SettingsBasedPlugin):
    DEFAULT_KEY = "tornado"

    def start(self, application: Application):
        settings = self.get_my_settings(application)
        debug = settings.get(DEBUG_KEY, False)
        serve_traceback = settings.get(SERVE_TRACEBACK_KEY, False)
        enable_pretty_logging()
        tornado = Tornado(debug=debug, serve_traceback=serve_traceback)

        return {
            "tornado": tornado,
            "http_server": HTTPServer(tornado),
        }



from alembic import context


def run_migrations(dbplugin, metadata):
    if context.is_offline_mode():
        run_migrations_offline(dbplugin, metadata)
    else:
        run_migrations_online(dbplugin, metadata)


def run_migrations_offline(dbplugin, metadata):
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        url=dbplugin.url,
        target_metadata=metadata,
        literal_binds=True,
    )

    run_alembic_migrations()


def run_migrations_online(dbplugin, metadata):
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = dbplugin.create_engine()

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=metadata)

        run_alembic_migrations()


def run_alembic_migrations():
    with context.begin_transaction():
        context.run_migrations()



class SettingMissing(Exception):
    """
    Setting's key is missing.
    """
    _FORMAT = ("'{0}' key is needed for use '{1}' database in application")

    def __init__(self, key, name):
        super().__init__()
        self.key = key
        self.name = name

    def __str__(self):
        return self._FORMAT.format(self.key, self.name)



from unittest.mock import MagicMock

from pytest import fixture
from pytest import raises

from qq import ApplicationNotStartedError
from qq import Context
from qq import SimpleInjector
from qq.application import Application
from qq.errors import InjectorNotInicialized
from qq.initializers import QQ_PARAMETER
from qq.initializers import ApplicationInitializer
from qq.initializers import get_initializers
from qq.initializers import get_injectors
from qq.initializers import is_initializer
from qq.initializers import is_injector
from qq.injector import Injector
from qq.plugins import SettingsPlugin

app = Application()


def not_injected_fun():
    pass


injected_fun = ApplicationInitializer(None)(not_injected_fun)


def example_fun(
    first, second, third=SimpleInjector(app, "key"), fourth=injected_fun
):
    return [first, second, third, fourth]


class TestIsInjector:
    @fixture
    def mparameter(self):
        return MagicMock()

    @fixture
    def mbound_args(self):
        args = MagicMock()
        args.arguments = {}
        return args

    def test_when_parameter_is_not_an_injector(self, mparameter, mbound_args):
        assert is_injector(mparameter, "key", mbound_args) is False

    def test_when_parameter_is_already_provided(self, mparameter, mbound_args):
        mbound_args.arguments["key"] = 1
        mparameter._default = Injector(None)

        assert is_injector(mparameter, "key", mbound_args) is False

    def test_when_parameter_is_not_provided(self, mparameter, mbound_args):
        mparameter._default = Injector(None)
        result = is_injector(mparameter, "key", mbound_args)
        assert result is True


class TestInjectors:
    def test_when_arguments_provided(self):
        assert list(get_injectors(example_fun, [1, 2, 3, 4], {})) == []

    def test_when_arguments_not_provided(self):
        """
        When injector argument not provided, injectors fun should return this
        injector.
        """
        context = Context(Application())
        context.values["key"] = 333

        name, value = list(get_injectors(example_fun, [1, 2], {}))[0]

        assert name == "third"
        assert value(context, "key") != value


class TestInjectedApplication:
    def test_when_arguments_provided(self):
        assert list(get_initializers(example_fun, [1, 2, 3, 4], {})) == []

    def test_when_arguments_not_provided(self):
        """
        When injector argument not provided, injectors fun should return this
        injector.
        """
        context = Context(Application())
        context.values["key"] = 333

        name, value = list(get_initializers(example_fun, [1, 2], {}))[0]

        assert name == "fourth"
        assert value == injected_fun


def default_settings():
    return {"settings": True}


class ExampleApplication(Application):
    def create_plugins(self):
        self.plugins["settings"] = SettingsPlugin("qq.tests.test_initializers")


class TestInitializeInjectors:
    @fixture
    def app(self):
        return ExampleApplication()

    @fixture
    def fun(self, app):
        def example_fun(
            first, second, third=SimpleInjector("settings"), fourth=injected_fun
        ):
            return [first, second, third, fourth]

        return ApplicationInitializer(app)(example_fun)

    @fixture
    def errorfun(self, app):
        def example_fun(
            first, second, third=SimpleInjector, fourth=injected_fun
        ):
            return [first, second, third, fourth]

        return ApplicationInitializer(app)(example_fun)

    @fixture
    def coroutine(self, app):
        async def example_fun(
            first, second, third=SimpleInjector("settings"), fourth=injected_fun
        ):
            return [first, second, third, fourth]

        return ApplicationInitializer(app)(example_fun)

    def test_when_arguments_provided(self, fun):
        assert fun(1, 2, 3, 4) == [1, 2, 3, 4]

    def test_when_arguments_not_provided_and_app_not_started(self, fun):
        with raises(ApplicationNotStartedError):
            fun(1, 2)

    def test_when_arguments_not_provided_and_app_started(self, app, fun):
        app.start("default_settings")
        result = fun(1, 2)
        assert result[0] == 1
        assert result[1] == 2
        assert result[2] == {"settings": True}
        assert getattr(result[3], QQ_PARAMETER) == app

    def test_swap_application(self, fun):
        secondfun = ApplicationInitializer(None)(fun)
        assert getattr(secondfun, QQ_PARAMETER) is None

    def test_when_not_initialized(self, app, errorfun):
        """
        Injected function should raise InjectorNotInicialized error when
        the Injector is wrongly used.
        """
        app.start("default_settings")
        with raises(InjectorNotInicialized):
            errorfun(1, 2)


class TestInitializeInjectorsForCoroutine:
    @fixture
    def app(self):
        return ExampleApplication()

    @fixture
    def fun(self, app):
        async def example_fun(
            first, second, third=SimpleInjector("settings"), fourth=injected_fun
        ):
            return [first, second, third, fourth]

        return ApplicationInitializer(app)(example_fun)

    @fixture
    def errorfun(self, app):
        async def example_fun(
            first, second, third=SimpleInjector, fourth=injected_fun
        ):
            return [first, second, third, fourth]

        return ApplicationInitializer(app)(example_fun)

    async def test_when_arguments_provided(self, fun):
        assert await fun(1, 2, 3, 4) == [1, 2, 3, 4]

    async def test_when_arguments_not_provided_and_app_not_started(self, fun):
        with raises(ApplicationNotStartedError):
            await fun(1, 2)

    async def test_when_arguments_not_provided_and_app_started(self, app, fun):
        app.start("default_settings")
        result = await fun(1, 2)
        assert result[0] == 1
        assert result[1] == 2
        assert result[2] == {"settings": True}
        assert getattr(result[3], QQ_PARAMETER) == app

    async def test_swap_application(self, fun):
        secondfun = ApplicationInitializer(None)(fun)
        assert getattr(secondfun, QQ_PARAMETER, None) is None

    async def test_when_not_initialized(self, app, errorfun):
        """
        Injected function should raise InjectorNotInicialized error when
        the Injector is wrongly used.
        """
        app.start("default_settings")
        with raises(InjectorNotInicialized):
            await errorfun(1, 2)



from contextlib import suppress
from sys import exc_info
from types import GeneratorType

from qq.application import Application
from qq.context import Context
from qq.errors import InjectorNotInicialized
from qq.initializers import ApplicationInitializer
from qq.types import BaseInjector


class Injector(BaseInjector):
    def __init__(self, fun):
        self.fun = fun
        self._initialized = False

    def __call__(self, *args, **kwargs):
        return self.__class__(self.fun).init(*args, **kwargs)

    def init(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self._initialized = True
        return self

    def start(self, application: Application):
        if not self._initialized:
            raise InjectorNotInicialized()
        self.context = Context(application)
        self.entered = self.context.__enter__()
        self.result = ApplicationInitializer(application)(self.fun)(
            self.entered, *self.args, **self.kwargs
        )
        if _is_generator(self.result):
            return self.result.__next__()
        else:
            return self.result

    def end(self):
        if _is_generator(self.result):
            with suppress(StopIteration):
                self.result.__next__()
        self.context.__exit__(*exc_info())


class ContextManagerInjector(Injector):
    def __init__(self, *args, **kwargs):
        super().__init__(None)
        self.init(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self  # pragma: no cover

    def start(self, application: Application):
        self.context = Context(application)
        self.entered = self.context.__enter__()
        return self.__enter__(self.entered, *self.args, **self.kwargs)

    def end(self):
        self.__exit__(
            *exc_info(),
            self.entered,
            *self.args,
            **self.kwargs,
        )
        self.context.__exit__(*exc_info())


@Injector
def SimpleInjector(context: Context, key: str):
    return context[key]


def _is_generator(obj):
    return isinstance(obj, GeneratorType)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['qq',
 'qq.plugins',
 'qq.plugins.celery',
 'qq.plugins.sqlalchemy',
 'qq.plugins.sqlalchemy.tests',
 'qq.plugins.tests',
 'qq.plugins.tornado',
 'qq.tests']

package_data = \
{'': ['*']}

install_requires = \
['python-dateutil>=2.8.2,<3.0.0']

extras_require = \
{'alembic': ['alembic>=1.0.9,<2.0.0'],
 'celery': ['celery>=5.2.2'],
 'developer': ['alembic>=1.0.9,<2.0.0',
               'sqlalchemy>=1.3.3,<2.0.0',
               'redis>=3.2.1,<4.0.0',
               'celery>=5.2.2'],
 'redis': ['redis>=3.2.1,<4.0.0'],
 'sqlalchemy': ['sqlalchemy>=1.3.3,<2.0.0']}

setup_kwargs = {
    'name': 'quackquack',
    'version': '1.2.0',
    'description': 'Quack Quack: A simple application framework',
    'long_description': 'About\n=====\n\n`Documentation <https://qqpy.org/>`_\n\n\nOverview\n--------\n\nThis project aims to resolve problem of configuring an application, which needs to\nhave initialization step (for example: for gathering settings or establishing\nconnections) and use Python style code (context managers and decorators) with\ndependency injection to get those data.\n\nFor example, normally you would need to use two separate mechanism for connection\nto the database (one for web, and one for celery). Mostly it uses the web framework\nconfiguration, to use in the celery code. It is fine, until a third sub-application\narrives. Or you have many microservices, where web frameworks are different\ndepending on the microservice purpose.\n\nSecond goal was to make synchronized code without any globals or magic. That is\nwhy using Quack Quack you know when the application is initialized (started),\nor where to look for code you are using.\n\nIn order to use QQ, you don\'t need to use hacks in some starting files, like\nimporting something from django, starting the application, and the import the\nrest.\n\nQuick Using Example\n-------------------\n\nTo use Quack Quack you need to create the application class (inherited from\n``qq.Application``\\ ) in which you need to add plugins. After configuring, you\nneed to "start" (initialize)\nthe application. After that you can use the application as context manager.\nAlso, you can make simple decorator, so you can use injectors (dependency\ninjection) in function\'s arguments.\n\n.. code-block:: python\n\n    from qq import Application\n    from qq import ApplicationInitializer\n    from qq import Context\n    from qq import SimpleInjector\n    from qq.plugins import SettingsPlugin\n    from qq.plugins.types import Settings\n\n\n    class MyApplication(Application):\n        def create_plugins(self):\n            self.plugins["settings"] = SettingsPlugin("settings")\n\n\n    application = MyApplication()\n    application.start("application")\n\n    with Context(application) as ctx:\n        print(ctx["settings"])\n\n    app = ApplicationInitializer(application)\n\n\n    @app\n    def samplefun(settings: Settings = SimpleInjector("settings")):\n        print(settings)\n\n\n    samplefun()\n    samplefun({"info": "fake settings"})  # dependency injection !!\n\n\n``context["settings"]`` in above example, is a variable made by the SettingsPlugin.\nIf you would like to know more, please go to the `Tutorial <docs/tutorial.md>`_\n\nInstallation\n------------\n\n.. code-block:: bash\n\n   pip install quackquack\n',
    'author': 'Dominik Dlugajczyk',
    'author_email': 'msocek@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/socek/quackquack',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from unittest.mock import MagicMock
from unittest.mock import call

from pytest import fixture

from qq import application
from qq.application import Application
from qq.finder import ObjectFinder
from qq.finder import is_defined_in


class TestIsDefinedIn:
    def test_when_true(self):
        assert is_defined_in(Application, application)

    def test_when_false(self):
        assert is_defined_in(is_defined_in, application) is False


class SampleFinder(ObjectFinder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.process_cache = {}

    def is_collectable(self, element):
        return getattr(element, "is_collectable") or False


class TestObjectFinder:
    @fixture
    def finder(self):
        return SampleFinder(["parent.one", "parent.two"], ["parent.one.arg1"])

    @fixture
    def mfind(self, mocker, finder):
        return mocker.patch.object(finder, "_find")

    @fixture
    def mget_all_packages(self, mocker, finder):
        return mocker.patch.object(finder, "_get_all_packages")

    @fixture
    def mfind_in_package(self, mocker, finder):
        return mocker.patch.object(finder, "_find_in_package")

    @fixture
    def mimport_module(self, mocker):
        return mocker.patch("qq.finder.import_module")

    @fixture
    def mwalk_packages(self, mocker):
        return mocker.patch("qq.finder.walk_packages")

    @fixture
    def module(self, mwalk_packages):
        module = MagicMock()
        mwalk_packages.return_value = [module]
        return module

    def test_find_when_cache_not_exists(self, finder, mfind):
        """
        .find should start find procedure if the cache for this class is not
        exsiting.
        """
        assert finder.find() == mfind.return_value
        mfind.assert_called_once_with()

    def test_find_when_cache_exists(self, finder, mfind):
        """
        .find should not start find procedure if the cache for this class is not
        exsiting.
        """
        finder.process_cache[finder.cache_key] = 123
        assert finder.find() == 123
        assert not mfind.called

    def test_find(self, finder, mget_all_packages, mfind_in_package):
        """
        ._find should get all packages and list add all elements from there.
        """
        mget_all_packages.return_value = [1]
        mfind_in_package.return_value = [2]

        assert finder._find() == [2, 2]

    def test_get_all_packages_when_import_failed(
        self, finder, mimport_module, mwalk_packages
    ):
        """
        ._get_all_packages should skip if import failed
        """
        mimport_module.side_effect = RuntimeError()

        list(finder._get_all_packages("parent"))

        mimport_module.assert_called_once_with(
            "parent",
        )
        assert not mwalk_packages.called

    def test_get_all_packages_when_module_in_ignore_list(
        self, finder, mimport_module, mwalk_packages, module
    ):
        """
        ._get_all_packages should skip if module in ignore list
        """
        module.name = "parent.one.arg1"
        mimport_module.return_value.__name__ = "prefixpkg"
        mimport_module.return_value.__path__ = "prefixpkg"

        list(finder._get_all_packages("parent"))

        mimport_module.assert_called_once_with(
            "parent",
        )
        mwalk_packages.assert_called_once_with("prefixpkg", "prefixpkg.")

    def test_get_all_packages_when_module_import_failed(
        self, finder, mimport_module, mwalk_packages, module
    ):
        """
        ._get_all_packages should skip packages that failed to import
        """
        module.name = "parent.one"
        parent = MagicMock()
        mimport_module.side_effect = [parent, RuntimeError()]
        parent.__name__ = "prefixpkg"
        parent.__path__ = "prefixpkg"

        list(finder._get_all_packages("parent"))

        mwalk_packages.assert_called_once_with("prefixpkg", "prefixpkg.")
        assert mimport_module.call_args_list == [
            call("parent"),
            call("parent.one"),
        ]

    def test_get_all_packages_when_module_import_success(
        self, finder, mimport_module, mwalk_packages, module
    ):
        """
        ._get_all_packages should return list of all collected modules
        """
        module.name = "parent.one"
        parent = MagicMock()
        child = MagicMock()
        mimport_module.side_effect = [parent, child]
        parent.__name__ = "prefixpkg"
        parent.__path__ = "prefixpkg"

        assert list(finder._get_all_packages("parent")) == [child]

        mwalk_packages.assert_called_once_with("prefixpkg", "prefixpkg.")
        assert mimport_module.call_args_list == [
            call("parent"),
            call("parent.one"),
        ]



from unittest.mock import sentinel

from pytest import fixture

from qq.plugins.logging import LoggingPlugin
from qq.testing import PluginFixtures

PREFIX = "qq.plugins.logging"


class TestLoggingPlugin(PluginFixtures):
    @fixture
    def plugin(self):
        plugin = LoggingPlugin()
        plugin.init("logging")
        return plugin

    @fixture
    def mdict_config(self, mocker):
        return mocker.patch(f"{PREFIX}.dictConfig")

    @fixture
    def mget_my_settings(self, mocker, plugin):
        mock = mocker.patch.object(plugin, "get_my_settings")
        mock.return_value = sentinel.logging
        return mock

    def test_start(self, plugin, mconfigurator, mdict_config, mget_my_settings):
        """
        .start should configure logging using settings
        """
        plugin.start(mconfigurator)

        mdict_config.assert_called_once_with(sentinel.logging)
